(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.genUnit = {}));
})(this, (function (exports) { 'use strict';

  const MICRO = '\xb5';

  function error(message) {
    return new TypeError(message);
  }
  function rangeError(message) {
    return new RangeError(message);
  }
  function errorInvalidOption(option) {
    return error("Invalid \"" + option + "\" option");
  }
  function errorOptionRemoved(removed, replacement) {
    return error("Option \"" + removed + "\" has been removed. Use \"" + replacement + "\" option instead.");
  }

  function isNullish(value) {
    return value == null;
  }
  function isType(type) {
    return value => {
      return typeof value === type;
    };
  }
  const isNumber = isType('number');
  const isObject = isType('object');
  const isFunction = isType('function');
  const {
    isArray
  } = Array;
  const {
    isFinite: isFiniteNumber
  } = Number;

  function transformFindItems(items, base) {
    return items.map(_ref => {
      let {
        pre,
        exp
      } = _ref;
      // compute multiplier
      const mul = base ** exp;
      // throw if multiplier is invalid
      if (!isFiniteNumber(mul) || mul <= 0) throw rangeError(base + " to the power of " + exp + " is not a valid multiplier");
      // return multiplier item
      return {
        pre,
        mul
      };
    });
  }

  /** @deprecated */
  // eslint-disable-next-line @typescript-eslint/no-deprecated
  function deprecated_transformAdvancedOptions(find) {
    // return option if it's a modern option
    if ('items' in find || !('find' in find)) return find;
    // handle deprecated option
    // use deprecated 'find' member as modern 'items' member
    // return modern option based on deprecated option
    // eslint-disable-next-line @typescript-eslint/no-deprecated
    const {
      find: items,
      base
    } = find;
    return {
      items,
      base
    };
  }
  function createFindTable(find, defaultItems, validateItems) {
    // return default table if "find" option is null or undefined
    if (isNullish(find)) return transformFindItems(defaultItems, 1000);
    // use find option as base if it's a number
    if (isNumber(find)) return transformFindItems(defaultItems, find);
    // use find option as items if it's an array
    if (isArray(find)) return transformFindItems(validateItems(find), 1000);
    // throw if find option is not an object at this point
    if (!isObject(find)) throw errorInvalidOption('find');
    // get items and base from find option
    // eslint-disable-next-line @typescript-eslint/no-deprecated
    const {
      items,
      base
    } = deprecated_transformAdvancedOptions(find);
    // normalize base
    const normalizedBase = base != null ? base : 1000;
    // return default items with given base if no items provided
    if (isNullish(items)) return transformFindItems(defaultItems, normalizedBase);
    // TODO: Throw more descriptive error
    // throw if items is not an array at this point
    if (!isArray(items)) throw errorInvalidOption('find');
    // return items based on option
    return transformFindItems(validateItems(items), normalizedBase);
  }

  const exa = {
    pre: 'E',
    exp: 6
  };
  const peta = {
    pre: 'P',
    exp: 5
  };
  const tera = {
    pre: 'T',
    exp: 4
  };
  const giga = {
    pre: 'G',
    exp: 3
  };
  const mega = {
    pre: 'M',
    exp: 2
  };
  const kilo = {
    pre: 'k',
    exp: 1
  };
  const milli = {
    pre: 'm',
    exp: -1
  };
  const micro = {
    pre: MICRO,
    exp: -2
  };
  const nano = {
    pre: 'n',
    exp: -3
  };
  const pico = {
    pre: 'p',
    exp: -4
  };
  const femto = {
    pre: 'f',
    exp: -5
  };
  const atto = {
    pre: 'a',
    exp: -6
  };

  function noDuplicates(items, key, str) {
    const {
      length
    } = items;
    for (let i1 = 0; i1 < length - 1; i1++) {
      const iut = items[i1][key];
      for (let i2 = i1 + 1; i2 < length; i2++) {
        if (items[i2][key] === iut) throw error("Duplicated " + str + " (" + iut + ")");
      }
    }
    return items;
  }

  function validateFormatItems(items) {
    // validate items
    // will throw on duplicated exponent
    const validated = noDuplicates(items, 'exp', 'exponent');
    // return sorted items
    return validated.sort((a, b) => b.exp - a.exp);
  }

  const defaultBase1000FormatExpItems = validateFormatItems([exa, peta, tera, giga, mega, kilo, {
    exp: 0,
    pre: ''
  }, milli, micro, nano, pico, femto, atto]);

  /** @deprecated */
  // eslint-disable-next-line @typescript-eslint/no-deprecated
  function deprecated_handleResult(item) {
    // return item if it's modern item
    if ('mul' in item) return item;
    // use deprecated 'div' member as modern 'mul' member
    // return modern item based on deprecated item
    // eslint-disable-next-line @typescript-eslint/no-deprecated
    const {
      pre,
      div: mul
    } = item;
    return {
      pre,
      mul
    };
  }
  function createUnitFinder(find) {
    // return wrapped user function if option it's a function
    // wrap function to test for result validity
    if (isFunction(find)) return value => {
      // call user function
      const result = find(value);
      // return undefined if user function returns nullish
      if (isNullish(result)) return;
      // throw if user function return non-object
      if (!isObject(result)) throw error(result + " is not a valid return value for \"find\" option");
      // normalize result for deprecated result
      // get prefix and multiplier
      // eslint-disable-next-line @typescript-eslint/no-deprecated
      const {
        pre,
        mul
      } = deprecated_handleResult(result);
      // throw if multiplier is invalid
      if (!isNumber(mul) || !isFiniteNumber(mul) || mul <= 0) throw rangeError(mul + " is not a valid multiplier");
      // return prefix and multiplier
      return {
        pre,
        mul
      };
    };
    // create find table
    const findTable = createFindTable(find, defaultBase1000FormatExpItems, validateFormatItems);
    // get table array length
    const {
      length: itemsLength
    } = findTable;
    // always return null if no table items
    if (itemsLength === 0) return () => null;
    return value => {
      // return undefined if value equals zero
      if (value === 0) return;
      // get absolute value
      const absolute = Math.abs(value);
      // iterate through table skipping last item (smallest unit)
      const lastIndex = itemsLength - 1;
      for (let i = 0; i < lastIndex; i++) {
        const item = findTable[i];
        if (absolute >= item.mul) return item;
      }
      // return smallest unit if it couldn't be found
      return findTable[lastIndex];
    };
  }

  function createFormatterFromOptions(space) {
    return (value, pre, unit) => {
      const wholeUnit = "" + pre + unit;
      const spacedUnit = wholeUnit ? "" + space + wholeUnit : '';
      return "" + value + spacedUnit;
    };
  }
  const oneSpace = ' ';
  function createOutputFormatter(output) {
    // return default formatter
    if (isNullish(output)) return createFormatterFromOptions(oneSpace);
    // return wrapped function if it's a function to ensure it returns a string
    if (isFunction(output)) {
      return (value, pre, unit) => {
        // call user function
        const formatted = output(value, pre, unit);
        // return user function result as string
        return "" + formatted;
      };
    }
    // throw if option is not an object at this point
    if (!isObject(output)) throw errorInvalidOption('output');
    // get option members
    const {
      space
    } = output;
    // if space member is a number
    if (isNumber(space)) {
      // throw if it's invalid number of spaces
      if (!isFiniteNumber(space) || space < 0) throw rangeError("Can't format output with " + space + " spaces");
      // return formatter with given number of spaces
      const spaceString = oneSpace.repeat(space);
      return createFormatterFromOptions(spaceString);
    }
    // normalize space
    const normalizedSpace = space != null ? space : oneSpace;
    // return formatter base on advanced options
    return createFormatterFromOptions(normalizedSpace);
  }

  function validateNumberOfDecimals(dec) {
    // throw if number of decimals is invalid
    if (!isFiniteNumber(dec) || dec < 0) throw rangeError("Can't create round function with " + dec + " decimal.");
    // return valid number of decimals
    return dec;
  }
  function createRounderFromOptions(dec, fixed) {
    // return fixed rounder function if fixed member set
    if (fixed) return num => {
      return num.toFixed(dec);
    };
    // compute round multiplier
    const roundMultiplier = 10 ** dec;
    // return regular rounder function
    return num => {
      return Math.round(num * roundMultiplier) / roundMultiplier;
    };
  }
  function createRounder(round) {
    // return default rounder if round option is nullish
    if (isNullish(round)) return createRounderFromOptions(2);
    // return user option if it's a function
    if (isFunction(round)) return round;
    // return rounder with round option as number of decimals if it's a number
    if (isNumber(round)) return createRounderFromOptions(validateNumberOfDecimals(round));
    // throw if round option is not an object at this point
    if (!isObject(round)) throw errorInvalidOption('round');
    // get members from advanced round option
    const {
      dec,
      fixed
    } = round;
    if (isNullish(dec)) return createRounderFromOptions(2, fixed);
    // TODO: throw more descriptive error
    // throw if dec member is not a number
    if (!isNumber(dec)) throw errorInvalidOption('round');
    // return rounder based on advanced options
    return createRounderFromOptions(validateNumberOfDecimals(dec), fixed);
  }

  /** @deprecated */
  // eslint-disable-next-line @typescript-eslint/no-deprecated
  function deprecated_createGetUnit(unit) {
    // return unit option if it's a function (deprecated)
    if (isFunction(unit)) return unit;
    // return a function returning normalized unit
    const normalizedUnit = unit != null ? unit : '';
    return () => normalizedUnit;
  }
  function createFormatter(options) {
    if (options === void 0) {
      options = {};
    }
    // TODO: remove in the future
    // option removed in version 0.1.0
    // throw if removed table option present
    if ('table' in options) throw errorOptionRemoved('table', 'find');
    // TODO: remove in the future
    // option removed in version 0.1.0
    // throw if removed dec option present
    if ('dec' in options) throw errorOptionRemoved('dec', 'round');
    // TODO: remove in the future
    // option removed in version 0.1.0
    // throw if removed fixed option present
    if ('fixed' in options) throw errorOptionRemoved('fixed', 'round');
    const {
      unit,
      find,
      round,
      output
    } = options;
    // eslint-disable-next-line @typescript-eslint/no-deprecated
    const getUnit = deprecated_createGetUnit(unit);
    const findUnit = createUnitFinder(find);
    const roundNum = createRounder(round);
    const formatOutput = createOutputFormatter(output);
    const formatWithPre = (value, pre) => {
      const rounded = roundNum(value);
      const computedUnit = getUnit(value, rounded, pre);
      return formatOutput(rounded, pre, computedUnit);
    };
    return value => {
      // return value as string if it's not finite
      if (!isFiniteNumber(value)) return "" + value;
      // find unit
      const item = findUnit(value);
      // return formatted value * 1 if can't find prefix and multiplier
      if (isNullish(item)) return formatWithPre(value, '');
      // get prefix and multiplier
      const {
        pre,
        mul: divisor
      } = item;
      // return formatted value
      return formatWithPre(value / divisor, pre);
    };
  }

  function format(value, options) {
    const formatter = createFormatter(options);
    return formatter(value);
  }

  function createExtractPre(unit) {
    if (unit) return wholeUnit => {
      // return prefix if whole unit ends with unit
      if (wholeUnit.endsWith(unit)) {
        const {
          length: unitLength
        } = unit;
        return wholeUnit.slice(0, -unitLength);
      }
      // return whole unit as prefix
      return wholeUnit;
    };
    // return whole unit as prefix
    return wholeUnit => wholeUnit;
  }

  function validateParseItems(items) {
    // validate items
    // will throw on duplicated prefix
    return noDuplicates(items, 'pre', 'prefix');
  }

  const defaultBase1000ParseExpItems = validateParseItems([exa, peta, tera, giga, {
    pre: 'meg',
    exp: 2
  }, mega, {
    pre: 'K',
    exp: 1
  }, kilo, milli, {
    pre: 'u',
    exp: -2
  }, micro, nano, pico, femto, atto]);

  function createMulFinder(find) {
    // return wrapped function if find option is a function
    // wrap function test for result validity
    if (isFunction(find)) return (prefix, unit) => {
      // find multiplier
      const result = find(prefix, unit);
      // return undefined if no multiplier found (null | undefined)
      if (isNullish(result)) return;
      // TODO: remove if future version
      // this feature was removed in version 0.1.0
      // throw error for removed feature
      if (isObject(result)) throw error('Function returning object is no longer supported, return a non-zero positive number, null or undefined.');
      // throw if multiplier is not valid
      if (!isNumber(result) || !isFiniteNumber(result) || result <= 0) throw rangeError(result + " is not a valid multiplier");
      // return multiplier
      return result;
    };
    // create find table
    const findTable = createFindTable(find, defaultBase1000ParseExpItems, validateParseItems);
    // return function returning null if there are no items to search
    if (findTable.length == 0) return () => null;
    // return multiplier finder function
    return prefix => {
      // return 1 as multiplier if captured prefix is empty
      if (!prefix) return 1;
      const item = findTable.find(_ref => {
        let {
          pre
        } = _ref;
        return pre === prefix;
      });
      // return undefined if not multiplier found
      if (!item) return;
      // return multiplier
      return item.mul;
    };
  }

  function createMatcher(match) {
    // return wrapped function if match option is a function
    // wrap function to test for result validity
    if (isFunction(match)) return input => {
      // call user function
      const captured = match(input);
      // return undefined if nullish return by user function
      if (isNullish(captured)) {
        return;
      }
      // throw if it's not an array
      if (!isArray(captured)) {
        throw error('Match function should return an array of strings');
      }
      // return array
      // array validation will happen later
      return captured;
    };
    // create regular expression based on option
    const re = isNullish(match) ? /^\s*(-?\d*\.?\d*(?:e[+-]?\d+)?)\s*([a-z\xb5]*)\s*$/i : new RegExp(match);
    // return input match function
    return input => {
      // execute RegExp against input
      const result = re.exec(input);
      // return undefined if it doesn't match
      if (!result) return;
      // get a slice of the result array containing the captured groups only
      const resultAsArray = result.slice(1);
      // return array
      // array validation will happen later
      return resultAsArray;
    };
  }

  function createParser(options) {
    if (options === void 0) {
      options = {};
    }
    // TODO: remove in the future
    // option removed in version 0.1.0
    // throw if removed table option present
    if ('table' in options) throw errorOptionRemoved('table', 'find');
    const {
      unit,
      match,
      find
    } = options;
    const extractPre = createExtractPre(unit);
    const findMultiplier = createMulFinder(find);
    const matchInput = createMatcher(match);
    // return parser function
    return input => {
      // if input is of type number (number, NaN or Infinity)
      if (isNumber(input)) {
        // return the number if it's finite (number)
        if (isFiniteNumber(input)) return input;
        // return NaN otherwise (NaN or Infinity)
        return NaN;
      }
      // return NaN if input is falsy (false, '', null or undefined) or true
      if (!input || input === true) return NaN;
      // convert input (probably non-empty string or object) to string
      const inputAsString = "" + input;
      // return NaN if input as string is empty ('')
      if (!inputAsString) return NaN;
      // convert string to number
      const inputAsNumber = +inputAsString;
      // return number if string as number is finite (not NaN or Infinity)
      if (isFiniteNumber(inputAsNumber)) return inputAsNumber;
      // capture value and unit from string
      const captured = matchInput(inputAsString);
      // return NaN if can't capture value and unit
      if (!captured) return NaN;
      // get captured array length
      const {
        length
      } = captured;
      // throw if captured array has less that 2 items
      if (length < 2) throw error("Match result array should have 2 items, got " + length);
      // get value and unit from captured array
      const [capturedValueAsString, wholeUnit] = captured;
      // return NaN if captured value is empty string ('')
      if (!capturedValueAsString) return NaN;
      // convert captured value to number
      const capturedValueAsNumber = +capturedValueAsString;
      // return NaN if captured value as number is not finite
      if (!isFiniteNumber(capturedValueAsNumber)) return NaN;
      // extract prefix from whole unit
      const pre = extractPre(wholeUnit);
      // try to find find multiplier
      const multiplier = findMultiplier(pre, unit);
      // return NaN if can't find multiplier
      if (isNullish(multiplier)) return NaN;
      // return parsed value
      return capturedValueAsNumber * multiplier;
    };
  }

  function parse(input, options) {
    const parser = createParser(options);
    return parser(input);
  }

  exports.MICRO = MICRO;
  exports.createFormatter = createFormatter;
  exports.createParser = createParser;
  exports.format = format;
  exports.parse = parse;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=gen-unit.umd.js.map
